package sem;

public class SemAux {

// If TypeName is not present in Env, or is not a type, emit an error message.
   public static sym.TypeSy ExpectType (String TypeName, sym.Env Env, int Pos) {
      sym.Symbol Type = Env.locateByName(TypeName);
      if (Type == null /*Standard.NoSy*/) {
         aux.Error.SemId(Pos, "Identifier not declared", TypeName);
         return null /*Standard.NoType*/;
      } else if (!(Type instanceof sym.TypeSy)) {
         aux.Error.SemId(Pos, "Type identifier expected", TypeName);
         return null /*Standard.NoType*/;
      }
      return (sym.TypeSy)Type;
   }

   public static sym.TypeSy ExpectTypeKind (String TypeName, sym.Env Env, java.lang.Class kind, String kindString, int Pos) {
      sym.TypeSy type = ExpectType(TypeName, Env, Pos);
      if (type == null) return null;
      if (kind.isInstance(type)) return type;
      aux.Error.SemId(Pos, kindString + " type expected", TypeName);
      return null;
   }


public static sym.TypeSy CheckTypeAndKind (
   sym.Symbol Type, 
   sym.TypeSy[] ExpectedType, 
   java.lang.Class[] ExpectedKind, 
   String Msg, 
   int Pos){
   if (Type == null /*Standard.NoType*/) return null /*Standard.NoType*/;

   boolean OK1 = false;
   for(int i=0; i<ExpectedType.length; i++)
      if (Type == ExpectedType[i]) OK1 = true;

   boolean OK2 = false;
   for(int i=0; i<ExpectedKind.length; i++)
       if (ExpectedKind[i].isInstance(Type)) OK2 = true;

   if (! (OK1 || OK2)) {
     aux.Error.Sem(Pos, Msg + " type expected");
     return null /*Standard.NoType*/;
   }
   return (sym.TypeSy)Type;
}

public static sym.TypeSy CheckType  (
   sym.Symbol Type, 
   sym.TypeSy[] ExpectedType, 
   String Msg, 
   int Pos){
   return CheckTypeAndKind(Type, ExpectedType, new java.lang.Class[]{}, Msg, Pos);
}

public static sym.TypeSy CheckUnTypeKind (
   sym.Symbol Type, 
   java.lang.Class[] NotExpected, 
   String Msg, 
   int Pos){
   if (Type == null /*Standard.NoType*/) return null /*Standard.NoType*/;

   boolean ERROR = false;
   for(int i=0; i<NotExpected.length; i++)
      if (NotExpected[i].isInstance(Type)) ERROR = true;

   if (ERROR) {
     aux.Error.Sem(Pos, Msg + " type expected");
     return null /*Standard.NoType*/;
   }
   return (sym.TypeSy)Type;
}

// If (A is not an integer or real type, print an error message.;
// If ( A is null /*Standard.NoType*/, no message is generated.;
// If (there is no error, A is returned. Otherwise null /*Standard.NoType*/ ;
// is returned.;
// Pos is the current line number.;
public static sym.TypeSy CheckNumeric(sym.Symbol A, int Pos) {
   return CheckType (A, new sym.TypeSy[]{sym.Standard.IntType, sym.Standard.RealType}, "Numeric", Pos);
}

// If (A is not an integer, print an error message.;
// If ( A is null /*Standard.NoType*/, no message is generated.;
// If (there is no error, A is returned. Otherwise null /*Standard.NoType*/ ;
// is returned.;
// Pos is the current line number.;
public static sym.TypeSy CheckInteger(sym.Symbol A, int Pos) {
   return CheckType (A,  new sym.TypeSy[]{sym.Standard.IntType}, "Integer", Pos);
}

// If (A is a record, array, or pointer type, print an error message.;
// If ( A is null /*Standard.NoType*/, no message is generated.;
// If (there is no error, A is returned. Otherwise null /*Standard.NoType*/ ;
// is returned.;
// Pos is the current line number.;
public static sym.TypeSy CheckScalar(sym.Symbol A, int Pos) {
   return CheckType (A, new sym.TypeSy[]{sym.Standard.IntType, sym.Standard.RealType, 
                                sym.Standard.BoolType, sym.Standard.CharType}, "Scalar", Pos);
}

public static sym.TypeSy CheckReadable(sym.Symbol A, int Pos) {
   return CheckType (A, new sym.TypeSy[]{sym.Standard.IntType, sym.Standard.RealType,
                                sym.Standard.CharType}, 
                                "INTEGER, REAL, CHAR", Pos);
}

public static sym.TypeSy CheckWriteable(sym.Symbol A, int Pos) {
   return CheckType (A, new sym.TypeSy[]{sym.Standard.IntType, sym.Standard.RealType, 
                                sym.Standard.StringType, sym.Standard.CharType},
                                "INTEGER, REAL, CHAR, STRING", Pos);
}

public static sym.TypeSy CheckAssignable(sym.Symbol A, int Pos) {
   return CheckUnTypeKind (A, new java.lang.Class[]{sym.ArrayType.class,sym.RecordType.class}, "Scalar", Pos);
}

public static sym.TypeSy CheckBasic(sym.Symbol A, int Pos) {
   return CheckUnTypeKind (A, new java.lang.Class[]{sym.ArrayType.class,sym.RecordType.class, 
                                               sym.ClassType.class,sym.RefType.class,sym.EnumType.class}, "Basic", Pos);
}

public static sym.TypeSy CheckComparable(sym.Symbol A, int Pos) {
   return CheckTypeAndKind (A, new sym.TypeSy[]{sym.Standard.IntType, sym.Standard.RealType, 
                                sym.Standard.StringType, sym.Standard.CharType}, 
                                new java.lang.Class[]{sym.ClassType.class,sym.RefType.class,sym.EnumType.class},
                               "Scalar or reference", Pos);
}

// If (A is not a boolean, print an error message.;
// If ( A is null /*Standard.NoType*/, no message is generated.;
// If (there is no error, A is returned. Otherwise null /*Standard.NoType*/ ;
// is returned.;
// Pos is the current line number.;
public static sym.TypeSy CheckBool(sym.Symbol A, int Pos) {
   return CheckType (A, new sym.TypeSy[]{sym.Standard.BoolType}, "Boolean", Pos);
}

public static sym.TypeSy CheckBinaryBool(sym.Symbol A, sym.Symbol B, int Pos) {
   sym.TypeSy x = CheckBool(A, Pos);
   sym.TypeSy y = CheckBool(B, Pos);
   if (x == null /*Standard.NoType*/ || y == null /*Standard.NoType*/)
      return null /*Standard.NoType*/;
   else
      return sym.Standard.BoolType;
}

// If (A and B don't have the same type, print an error message.;
// If (either A or B are null /*Standard.NoType*/, no message is generated.;
// If (there is no error, C is returned. Otherwise null /*Standard.NoType*/ ;
// is returned.;
// Pos is the current line number.;
public static sym.TypeSy CheckBinary(sym.Symbol A, sym.Symbol B, sym.TypeSy C, int Pos) {
   if (A == B) return C;
   if (A instanceof sym.ClassType && B instanceof sym.ClassType && 
          (IsSuperType((sym.ClassType)A,(sym.ClassType)B) || 
           IsSuperType((sym.ClassType)B,(sym.ClassType)A)))
      return C;

   if (A != null /*Standard.NoType*/ && B != null /*Standard.NoType*/) 
      aux.Error.Sem(Pos, "Type missmatch");

   return null /*Standard.NoType*/;
}

public static boolean Assignable(sym.Symbol Left, sym.TypeSy Right) {
    if (Left == Right) return true;
    if (! ((Left instanceof sym.ClassType) && (Right instanceof sym.ClassType)))
       return false;
    sym.ClassType L = (sym.ClassType)Left;
    sym.ClassType R = (sym.ClassType)Right;
    if (IsSuperType(L,R) || IsSuperType(R,L))
       return true;
    return false;
}

//================================================;
public static sym.SyTab ExtractVariables(sym.SyTab D) {
   sym.SyTab S = new sym.SyTab();
   java.util.Iterator i = D.iterator();
   while (i.hasNext()) {
      sym.Symbol v = (sym.Symbol)i.next();
      if (v.getClass() == sym.VariableSy.class)
        S = S.insert(v);
   }
   return S;
}

//================================================;
public static sym.Symbol LocateClassField (sym.ClassType Class, String Name) {
   java.util.LinkedList superClasses = SuperTypes(Class);
   java.util.ListIterator i = superClasses.listIterator();
   while (i.hasNext()) {
      sym.ClassType c = (sym.ClassType) i.next();
      sym.SyTab fields = c.GetFields();
      sym.Symbol n = fields.locateByName(Name);
      if (n != null /*Standard.NoSy*/) return n;
   }
   return null /*Standard.NoSy*/;
}


public static sym.Env ConcatClassFields (sym.ClassType Class) {
   sym.Env e = new sym.Env();
   java.util.LinkedList superClasses = SuperTypes(Class);
   java.util.ListIterator i = superClasses.listIterator();
   while (i.hasNext()) {
      sym.ClassType c = (sym.ClassType) i.next();
      sym.SyTab fields = c.GetFields();
      e = e.snoc(fields);
   }
   return e;
}

public static java.util.LinkedList ProperSuperTypes (sym.ClassType Class) {
   java.util.LinkedList L = new java.util.LinkedList();
   if (Class != sym.Standard.RootSy)
      while (true) {
         Class = (sym.ClassType) Class.GetSuperType();
         if (Class == sym.Standard.RootSy) break;
         L.add(Class);
      }
   L.add(sym.Standard.RootSy);
   return L;
}

public static java.util.LinkedList SuperTypes (sym.ClassType Class) {
   java.util.LinkedList L = ProperSuperTypes(Class);
   L.addFirst(Class);
   return L;
}

public static boolean IsSuperType (sym.ClassType ParentClass, sym.ClassType ChildClass){
   java.util.LinkedList L = SuperTypes(ChildClass);
   java.util.ListIterator i = L.listIterator();
   while (i.hasNext()) {
      sym.ClassType c = (sym.ClassType) i.next();
      if (ParentClass == c) return true;
   }
   return false;
}

public static boolean IsProperSuperType (sym.ClassType ParentClass, sym.ClassType ChildClass){
   java.util.LinkedList L = ProperSuperTypes(ChildClass);
   java.util.ListIterator i = L.listIterator();
   while (i.hasNext()) {
      sym.ClassType c = (sym.ClassType) i.next();
      if (ParentClass == c) return true;
   }
   return false;
}

public static boolean CheckSignature(sym.ProcedureSy NewMethod, sym.ProcedureSy OldMethod) {
   sym.SyTab NFormals = NewMethod.GetProcFormals();
   sym.SyTab OFormals = OldMethod.GetProcFormals();
   int FormalNo = 1;
   while(true) {
       sym.FormalSy NFormal = sym.ProcedureSy.GetFormalParam(NFormals,FormalNo);
       sym.FormalSy OFormal = sym.ProcedureSy.GetFormalParam(OFormals,FormalNo);
      if (NFormal == null /*Standard.NoSy*/ && OFormal == null /*Standard.NoSy*/)
         return true; 
      if (NFormal == null /*Standard.NoSy*/ || OFormal == null /*Standard.NoSy*/)
         return false;

      sym.TypeSy NFormalType = NFormal.GetType();
      sym.TypeSy OFormalType = OFormal.GetType();

      if (! Assignable(OFormalType, NFormalType))
          return false;

      String NFormalMode = NFormal.GetFormalMode();
      String OFormalMode = OFormal.GetFormalMode();
      if (!NFormalMode.equals(OFormalMode))
         return false;

      FormalNo++;
   }
}

//================================================;
// Resolve forward references in ref types.;

static class RefForward {
   public sym.RefType Symbol;
   public String TypeName;
   public int Pos;
   public RefForward  (sym.RefType Symbol, String TypeName, int Pos) {
      this.Symbol=Symbol; this.TypeName = TypeName; this.Pos = Pos;
   }
}

static java.util.LinkedList RefTypes = new java.util.LinkedList();

public static void AddForwardRef (sym.RefType Symbol, String TypeName, int Pos) {
   RefTypes.add(new RefForward(Symbol, TypeName, Pos));
}

public static void ResolveForwards (sym.Env Env) {
   if (RefTypes==null) return;
   java.util.ListIterator i = RefTypes.listIterator();
   while (i.hasNext()) {
      RefForward t = (RefForward)i.next();
      sym.TypeSy Type = (sym.TypeSy)Env.locateByName(t.TypeName);
      if (Type == null /*Standard.NoSy*/) {
         aux.Error.SemId(t.Pos, "Identifier not declared", t.TypeName);
         Type = null /*Standard.NoType*/;
      } else
         t.Symbol.SetRefType(Type);
   }
   RefTypes = new java.util.LinkedList();
}


//================================================;

public static sym.Value EvalBinary(boolean IsConst, sym.Value L, sym.Value R, sym.TypeSy Type, int Op, int Pos) {
   sym.Value result = null;
   if (!IsConst) return null;

   if (Type == sym.Standard.StringType)
       result = sym.StringValue.binaryEval(Op, (sym.StringValue)L, (sym.StringValue)R, Pos);
   else if (Type == sym.Standard.IntType)
       result = sym.IntValue.binaryEval(Op, (sym.IntValue)L, (sym.IntValue)R, Pos);
   else if (Type == sym.Standard.RealType)
       result = sym.RealValue.binaryEval(Op, (sym.RealValue)L, (sym.RealValue)R, Pos);
   else if (Type == sym.Standard.BoolType)
       result = sym.BoolValue.binaryEval(Op, (sym.BoolValue)L, (sym.BoolValue)R, Pos);
   else if (Type == sym.Standard.CharType)
       result = sym.CharValue.binaryEval(Op, (sym.CharValue)L, (sym.CharValue)R, Pos);

   if (result instanceof sym.ErrorValue) {
       sym.ErrorValue e = (sym.ErrorValue) result;
       aux.Error.Sem(e.Pos, e.msg);
   }
   return result;
}

public static sym.Value EvalUnary(boolean IsConst, sym.Value L, sym.TypeSy Type, int Op, int Pos) {
   if (!IsConst) return null;
   sym.Value result = null;
   if ((Type == sym.Standard.IntType)) 
       result = sym.IntValue.unaryEval(Op, (sym.IntValue)L, Pos);
   else if ((Type == sym.Standard.RealType)) 
       result = sym.RealValue.unaryEval(Op,(sym.RealValue) L, Pos);
   else if ((Type == sym.Standard.BoolType)) 
       result = sym.BoolValue.unaryEval(Op, (sym.BoolValue)L, Pos);

   if (result instanceof sym.ErrorValue) {
       sym.ErrorValue e = (sym.ErrorValue) result;
       aux.Error.Sem(e.Pos, e.msg);
   }
  return result;
}


//===========================================================;
//public static java.util.LinkedList GenElementTypes(Type) {
//   TypeKind = symbol_GetTypeKind(Type);
//   if (TypeKind == "ArrayType");
//      susp} GenElementTypes(symbol_GetArrayElementType(Type));
//   else if (TypeKind == "RecordType") {
//      Fields = symbol_GetFields(Type);
//      every i = sytab_Ids(Fields) do;
//         susp} GenElementTypes(symbol_GetType(i));
//  } else
//     susp} Type;
//}

public static sym.TypeSy GetLargestElementType(sym.TypeSy Type) {
   return sym.Standard.IntType;
//   maxSize = 0;
//   maxType = null /*Standard.NoSy*/;
//   every i = GenElementTypes(Type) do;
//      if (symbol_GetSize(i) > maxSize) {
//         maxSize = symbol_GetSize(i);
//         maxType = i;
//     }
//   return maxType;
}


//===========================================================;
static int TempVarNo = 0;

public static sym.VariableSy MakeTempVar(
    sym.ProcedureSy CurrentProc, sym.TypeSy Type, int Level, int Pos) {
   sym.SyTab locals = CurrentProc.GetProcLocals();
   int localSize = CurrentProc.GetLocalSize();

   sym.VariableSy var = new sym.VariableSy("$" + TempVarNo, Pos, Level);
   TempVarNo++;
   var.SetType(Type);
   var.SetOffset(sym.Arch.CalcOffset(localSize, Type));
   var.SetSize(Type.GetSize());

   localSize = sym.Arch.AddAlign(localSize, Type);
   locals = locals.insert(var);
   CurrentProc.SetProcLocals(locals);
   CurrentProc.SetLocalSize(localSize);
   return var;
}

static int RegVarNo = 0;
public static sym.TempSy MakeTempReg(sym.TypeSy Type, int Pos) {
   sym.TempSy var = new sym.TempSy("@" + RegVarNo, Pos, -1);
   RegVarNo++;
   var.SetType(Type);
   var.SetOffset(-1);
   var.SetSize(Type.GetSize());
   return var;
}

static int TempConstNo = 0;
public static sym.ConstSy MakeTempConst(sym.TypeSy Type, sym.Value Value, int Pos) {
   sym.ConstSy c = new sym.ConstSy("%" + TempConstNo, Pos, -1);
   c.SetType(Type);
   TempConstNo++;
   c.SetSize(Type.GetSize());
   c.SetConstantValue(Value);
   return c;
}

//===========================================================;
public static void LargeValueFormalCopies(
   sym.ProcedureSy CurrentProc, 
   sym.SyTab Formals, 
   int Pos) {
   java.util.Iterator i = Formals.iterator();
   while (i.hasNext()) {
      sym.FormalSy f = (sym.FormalSy) i.next();
      int size = f.GetSize();
      String mode = f.GetFormalMode();
      if ((mode.equals("VAL") && (size > sym.Standard.IntType.GetSize()))) {
         sym.TypeSy Type = f.GetType();
         int level = f.GetLevel();
         sym.VariableSy v = MakeTempVar(CurrentProc, Type, level, Pos);
         f.SetFormalCopy(v);
      }
   }
}

}
